# Copyright (C) 2010  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import grp
import hashlib
import os
import pwd
import stat
import time
import urllib


class NumberNormalizer(object):

    '''Normalize inode and device numbers.
    
    When we make two manifests of the same directory tree, but the
    tree may have been moved to another disk, the inode and device
    numbers may be different. This should not be a cause for concern,
    however. What is important is that if two names were hardlinked
    to the same file before, they still are, and if they weren't,
    they still aren't.
    
    To achieve this, we normalize the inode and device numbers.
    The input files are fed to the normalizer in a deterministic
    sequence, and the sequence defines the numbers we use. Thus,
    if the input files have inode numbers [42, 13, 105], we produce
    [1, 2, 3]. If one of the input numbers is repeated, that number
    is re-used.
    
    This is not a perfect solution. If the second manifest has a
    new file, it will throw off the entire remaining sequence, causing
    a big diff. But we'll live with that.
    
    '''
    
    def __init__(self):
        self.reset()

    def get(self, input_number, numbers, next):
        if input_number in numbers:
            return numbers[input_number], next
        else:
            numbers[input_number] = next
            return numbers[input_number], next + 1
    
    def get_ino(self, ino):
        output, self.next_ino = self.get(ino, self.ino_numbers, self.next_ino)
        return output
    
    def get_dev(self, dev):
        output, self.next_dev = self.get(dev, self.dev_numbers, self.next_dev)
        return output

    def reset(self):
        '''This is used by unit tests.'''
        self.ino_numbers = dict()
        self.next_ino = 1

        self.dev_numbers = dict()
        self.next_dev = 1


class FilesystemObject(object):

    '''An object in the file system.
    
    Responsible for gathering information and formatting it for
    reporting.
    
    The optional arguments are intended for unit tests.
    
    '''
    
    def __init__(self, filename, normalizer, stat_result=None, sha1=None,
                 open_file=None, readlink=None):
        stat_result = stat_result or os.lstat(filename)
        sha1 = sha1 or hashlib.sha1()
        self.open_file = open_file or file
        readlink = readlink or os.readlink
        
        self.values = dict()
        self['Name'] = filename
        self['Mtime'] = self.format_time(stat_result.st_mtime)
        self['Mode'] = '%o' % stat_result.st_mode
        self['Ino'] = '%d' % normalizer.get_ino(stat_result.st_ino)
        self['Dev'] = '%d' % normalizer.get_dev(stat_result.st_dev)
        self['Nlink'] = '%d' % stat_result.st_nlink
        self['Size'] = '%d' % stat_result.st_size
        self['Uid'] = '%d' % stat_result.st_uid
        self['Username'] = self.lookup_username(stat_result.st_uid)
        self['Gid'] = '%d' % stat_result.st_gid
        self['Group'] = self.lookup_group(stat_result.st_gid)
        if stat.S_ISREG(stat_result.st_mode):
            self['Sha-1'] = self.compute_sha1(filename, sha1)
        if stat.S_ISLNK(stat_result.st_mode):
            self['Target'] = readlink(filename)

    def format_time(self, timestamp):
        return time.strftime('%Y-%m-%d %H:%M:%S +0000', 
                             time.gmtime(timestamp))

    def lookup_username(self, uid):
        return pwd.getpwuid(uid).pw_name

    def lookup_group(self, gid):
        return grp.getgrgid(gid).gr_name

    def compute_sha1(self, filename, sha1):
        f = self.open_file(filename)
        while True:
            data = f.read(64*1024) # 64 KiB seems reasonable.
            if not data:
                break
            sha1.update(data)
        f.close()
        return sha1.hexdigest()

    def hook_name(self, value):
        return urllib.quote(value)

    def __setitem__(self, key, value):
        hook_name = 'hook_%s' % key.lower()
        if hasattr(self, hook_name):
            value = getattr(self, hook_name)(value)
        self.values[key] = value
        
    def __getitem__(self, key):
        return self.values.get(key, '')
        
    def format(self): # pragma: no cover
        keys = ['Name'] + [x 
                           for x in sorted(self.values.keys())
                           if x != 'Name']
        return ''.join('%s: %s\n' % (key, self[key]) for key in keys)
